#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <err.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/mman.h>

#define UNUSED		__attribute__ ((unused))

volatile int end;

void sighandler1(UNUSED int signo)
{
}

void sighandler2(UNUSED int signo)
{
	end = 1;
}

int mem_hog(void)
{
	sigset_t sigset;
	long pagesize;
	unsigned long *addr;
	int ret;

	pagesize = getpagesize();

	if (sigemptyset(&sigset) < 0)
		err(1, "sigemptyset()");
	sigsuspend(&sigset);

	while (!end) {
		addr = mmap(NULL, pagesize * 10, PROT_READ | PROT_WRITE,
				MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
		if (addr == MAP_FAILED) {
			ret = 1;
			perror("mmap failed.\n");
			break;
		}
		memset(addr, 0xF7, pagesize * 10);
		munmap(addr, pagesize * 10);
	}

	return ret;
}

int main(int argc, char **argv)
{
	int ntasks;
	int i;
	int ret = 0;
	int status;
	struct sigaction sa1, sa2;
	pid_t *pids;

	if (argc != 2)
		err(1, "option is wrong.");

	ntasks = atoi(argv[1]);

	if (ntasks <= 0)
		err(1, "ntasks is small.");

	pids = malloc(ntasks * sizeof(pid_t));
	if (!pids)
		err(1, "allocate pids failed.");

	sa1.sa_handler = sighandler1;
	if (sigemptyset(&sa1.sa_mask) < 0)
		err(1, "sigemptyset()");

	sa1.sa_flags = 0;
	if (sigaction(SIGUSR1, &sa1, NULL) < 0)
		err(1, "sigaction()");

	sa2.sa_handler = sighandler2;
	if (sigemptyset(&sa2.sa_mask) < 0)
		err(1, "sigemptyset()");

	sa2.sa_flags = 0;
	if (sigaction(SIGUSR2, &sa2, NULL) < 0)
		err(1, "sigaction()");

	for (i = 0; i < ntasks; i++) {
		pids[i] = fork();
		if (pids[i] < 0) {
			perror("fork() failed.\n");
			ret = 1;
			break;
		} else if (!pids[i])
			return mem_hog();
	}

	if (i != ntasks) {
		ntasks = i;
		while (--i >= 0)
			kill(pids[i], SIGUSR2);
	}

	while (ntasks--) {
		wait(&status);
		if (WIFEXITED(status) && WEXITSTATUS(status))
			ret = 1;
	}

	return ret;
}
